
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

public class Cell {
	Data data;
	HashMap<ArrayList<Integer>,Info> candidates;
	HashMap<ArrayList<Integer>,HashMap<Integer,Object>> nextCands;
	boolean allLessCorThr;
	
	public Cell(Data data){
		candidates = new HashMap<ArrayList<Integer>,Info>();
		this.data = data;
		this.allLessCorThr = true;
	}
	
	void computeCandsFromCands(int k){
		if(data.pruningType==CorMiner.ADVANCED_PRUNING){
			data.initOneItemsCor();
		}

		//Go through entire input data
		for(HashMap<Integer,Object> tran:data.trans){
			Set<Integer> row = tran.keySet();
			Set<ArrayList<Integer>> keys = candidates.keySet();
			for(ArrayList<Integer> key:keys){
				if(row.containsAll(key)){
					Info info = candidates.get(key);
					info.sup++;
					candidates.put(key, info);
				}
			}
		}
		
		updateCands(k);
	}
	
	void updateCands(int k){
		//Apply minsup
		Set<ArrayList<Integer>> keys = candidates.keySet();
		HashMap<ArrayList<Integer>,Object> toRemove = new HashMap<ArrayList<Integer>,Object>();
		for(ArrayList<Integer> key:keys){
			Info info = candidates.get(key);
			if(info.sup < data.minsup) toRemove.put(key,null);
			else { //compute correlation value
				double oneSups = 1;
				for(int i:key){
					oneSups = oneSups * data.oneItemSets.get(i).sup;
				}
				//info.cor = (double)info.sup/Math.sqrt(oneSups);
				info.cor = (double)info.sup/Math.pow(oneSups,(double)1/k);
				candidates.put(key, info);

				if(info.cor >= data.corThr){
					allLessCorThr = false;
					//System.out.println(key + " " + info.cor);
					data.sols.put(new ArrayList<Integer>(key), info);
				}
				if(data.pruningType==CorMiner.ADVANCED_PRUNING){
					for(int i:key){
						Info oneItemInfo = data.oneItemSets.get(i);
						if(oneItemInfo.cor < info.cor){
							oneItemInfo.cor = info.cor;
							data.oneItemSets.put(i, oneItemInfo);
						}
//						double cp = (double)info.sup/oneItemInfo.sup;
//						if(oneItemInfo.cor < cp){
//							oneItemInfo.cor = cp;
//							data.oneItemSets.put(i, oneItemInfo);
//						}
					}
				}
			}
		}
		Set<ArrayList<Integer>> toRemoves = toRemove.keySet();
		for(ArrayList<Integer> key:toRemoves) candidates.remove(key);
	}
	
	
	
	void computeCandsFromTrans(int k){
		if(data.pruningType==CorMiner.ADVANCED_PRUNING){
			data.initOneItemsCor();
		}

		//Go through entire input data
		for(HashMap<Integer,Object> tran:data.trans){
			Set<Integer> row = tran.keySet();
			Integer[] rowArr = row.toArray(new Integer[0]);
			Arrays.sort(rowArr);
			ArrayList<Integer[]> kItems = generateKItems(k,rowArr);
			for(Integer[] kItem:kItems){
//				System.out.print(Arrays.asList(kItem));
				ArrayList<Integer> key = new ArrayList<Integer>(Arrays.asList(kItem));
				if(candidates.containsKey(key)){
					Info info = candidates.get(key);
					info.sup++;
					candidates.put(key, info);
				}
			}
//			System.out.println();
		}

		updateCands(k);
	}


	void computeCandsFromTransNew(int k){
		if(data.pruningType==CorMiner.ADVANCED_PRUNING){
			data.initOneItemsCor();
		}

		//Go through entire input data
		for(HashMap<Integer,Object> tran:data.trans){
			Set<Integer> row = tran.keySet();
//			Integer[] rowArr = row.toArray(new Integer[0]);
//			Arrays.sort(rowArr);
			ArrayList<Integer> rowArr = new ArrayList<Integer>(row);
			Collections.sort(rowArr);
			ArrayList<ArrayList<Integer>> kItems = generateKItemsNew(k,rowArr);
			for(ArrayList<Integer> key:kItems){
//				System.out.print(key);
				if(candidates.containsKey(key)){
					Info info = candidates.get(key);
					info.sup++;
					candidates.put(key, info);
				}
			}
//			System.out.println();
		}

		updateCands(k);
	}

	void computeCandsTopKFromCands(int k){
		if(data.pruningType==CorMiner.ADVANCED_PRUNING){
			data.initOneItemsCor();
		}

		//Go through entire input data
		for(HashMap<Integer,Object> tran:data.trans){
			Set<Integer> row = tran.keySet();
			Set<ArrayList<Integer>> keys = candidates.keySet();
			for(ArrayList<Integer> key:keys){
				if(row.containsAll(key)){
					Info info = candidates.get(key);
					info.sup++;
					candidates.put(key, info);
				}
			}
		}
		
		updateCandsTopK(k);
	}
	
	void updateCandsTopK(int k){
		Set<ArrayList<Integer>> keys = candidates.keySet();
		HashMap<ArrayList<Integer>,Object> toRemove = new HashMap<ArrayList<Integer>,Object>();
		for(ArrayList<Integer> key:keys){
			Info info = candidates.get(key);
			if(info.sup < data.minsup) toRemove.put(key,null);
			else { //compute correlation value
				double oneSups = 1;
				for(int i:key){
					oneSups = oneSups * data.oneItemSets.get(i).sup;
				}
				//info.cor = (double)info.sup/Math.sqrt(oneSups);
				info.cor = (double)info.sup/Math.pow(oneSups,(double)1/k);
				candidates.put(key, info);

				if(info.cor > data.corThr){
					allLessCorThr = false;
					if(data.topkCors.containsKey(info.cor)) {
						int num = data.topkCors.get(info.cor);
						data.topkCors.put(info.cor, num+1);
					} else {
						data.topkCors.put(info.cor,1);
					}
					if(data.topkInvs.containsKey(info.cor)){
						HashMap<ArrayList<Integer>,Object> vals = data.topkInvs.get(info.cor);
						vals.put(new ArrayList<Integer>(key), null);
						data.topkInvs.put(info.cor,vals);						
					} else {
						HashMap<ArrayList<Integer>,Object> vals = new HashMap<ArrayList<Integer>,Object>();
						vals.put(new ArrayList<Integer>(key), null);
						data.topkInvs.put(info.cor,vals);
					}
					data.sols.put(new ArrayList<Integer>(key), info);
					if(data.sols.size() > data.topK){
						Double removeVal = data.topkCors.firstKey();
						int count = data.topkCors.get(removeVal);
						if(count>1) {
							count--;
							data.topkCors.put(removeVal, count);
							HashMap<ArrayList<Integer>,Object> vals = data.topkInvs.get(removeVal);
							Iterator<ArrayList<Integer>> keyIterator = vals.keySet().iterator();
							//ArrayList<Integer> firstKey = new ArrayList<Integer>(keyIterator.next());
							ArrayList<Integer> firstKey = keyIterator.next();
							data.sols.remove(firstKey);
							vals.remove(firstKey);
							data.topkInvs.put(removeVal, vals);
						} else {
							data.topkCors.remove(removeVal);
							HashMap<ArrayList<Integer>,Object> vals = data.topkInvs.get(removeVal);
							Iterator<ArrayList<Integer>> keyIterator = vals.keySet().iterator();
							ArrayList<Integer> firstKey = keyIterator.next();
							data.sols.remove(firstKey);
							data.topkInvs.remove(removeVal);
						}
						data.corThr = data.topkCors.firstKey();
					}
				}
				if(data.pruningType==CorMiner.ADVANCED_PRUNING){
					for(int i:key){
						Info oneItemInfo = data.oneItemSets.get(i);
						if(oneItemInfo.cor < info.cor){
							oneItemInfo.cor = info.cor;
							data.oneItemSets.put(i, oneItemInfo);
						}
//						double cp = (double)info.sup/oneItemInfo.sup;
//						if(oneItemInfo.cor < cp){
//							oneItemInfo.cor = cp;
//							data.oneItemSets.put(i, oneItemInfo);
//						}
					}
				}
			}
		}
		Set<ArrayList<Integer>> toRemoves = toRemove.keySet();
		for(ArrayList<Integer> key:toRemoves) candidates.remove(key);
	}
	
	void computeCandsTopKFromTrans(int k){
		if(data.pruningType==CorMiner.ADVANCED_PRUNING){
			data.initOneItemsCor();
		}

		//Go through entire input data
		for(HashMap<Integer,Object> tran:data.trans){
			Set<Integer> row = tran.keySet();
			Integer[] rowArr = row.toArray(new Integer[0]);
			Arrays.sort(rowArr);
			ArrayList<Integer[]> kItems = generateKItems(k,rowArr);
			for(Integer[] kItem:kItems){
//				System.out.print(Arrays.asList(kItem));
				ArrayList<Integer> key = new ArrayList<Integer>(Arrays.asList(kItem));
				if(candidates.containsKey(key)){
					Info info = candidates.get(key);
					info.sup++;
					candidates.put(key, info);
				}
			}
//			System.out.println();
		}
		
		updateCandsTopK(k);
	}
	
	void computeCandsTopKFromTransNew(int k){
		if(data.pruningType==CorMiner.ADVANCED_PRUNING){
			data.initOneItemsCor();
		}

		//Go through entire input data
		for(HashMap<Integer,Object> tran:data.trans){
			Set<Integer> row = tran.keySet();
//			Integer[] rowArr = row.toArray(new Integer[0]);
//			Arrays.sort(rowArr);
			ArrayList<Integer> rowArr = new ArrayList<Integer>(row);
			Collections.sort(rowArr);
			ArrayList<ArrayList<Integer>> kItems = generateKItemsNew(k,rowArr);
			for(ArrayList<Integer> key:kItems){
//				System.out.print(key);
				if(candidates.containsKey(key)){
					Info info = candidates.get(key);
					info.sup++;
					candidates.put(key, info);
				}
			}
//			System.out.println();
		}
		
		updateCandsTopK(k);
	}
	
	ArrayList<Integer[]> generateKItems(int k, Integer[] row){
		ArrayList<Integer[]> kItems = new ArrayList<Integer[]>();
		int len = row.length;
		int max = len-k;
		for(int i=0;i<=max;i++){
			Integer[] cand = new Integer[k];
			cand[0] = row[i];
			generateKItems(k,row,len,cand,1,i+1,kItems);
		}
		
		return kItems;
	}
	
	void generateKItems(int k, Integer[] row, int len, Integer[] cand, int curSize, int nextIdx, ArrayList<Integer[]> kItems){
		if(k>curSize){
			if(k-curSize<=len-nextIdx){
				for(int i=nextIdx;i<len;i++){
					cand[curSize]=row[i];
					generateKItems(k,row,len,cand,curSize+1,i+1,kItems);
				}
			}
		}else{
			kItems.add(Arrays.copyOf(cand,k));
		}
	}
	
	ArrayList<ArrayList<Integer>> generateKItemsNew(int k, ArrayList<Integer> row){
		ArrayList<ArrayList<Integer>> kItems = new ArrayList<ArrayList<Integer>>();
		int len = row.size();
		int max = len-k;
		for(int i=0;i<=max;i++){
			ArrayList<Integer> cand = new ArrayList<Integer>();
			for(int j=0;j<k;j++) cand.add(-1);
			cand.set(0, row.get(i));
			generateKItemsNew(k,row,len,cand,1,i+1,kItems);
		}
		
		return kItems;
	}
	
	void generateKItemsNew(int k, ArrayList<Integer> row, int len, ArrayList<Integer> cand, int curSize, int nextIdx, ArrayList<ArrayList<Integer>> kItems){
		if(k>curSize){
			if(k-curSize<=len-nextIdx){
				for(int i=nextIdx;i<len;i++){
					cand.set(curSize, row.get(i));
					generateKItemsNew(k,row,len,cand,curSize+1,i+1,kItems);
				}
			}
		}else{
			kItems.add(new ArrayList<Integer>(cand));
		}
	}
	
	void prepareNextCands(){
		nextCands = new HashMap<ArrayList<Integer>,HashMap<Integer,Object>>();
		if(data.pruningType==CorMiner.ADVANCED_PRUNING) {
			data.pruneOneItemsCorSorted();
			//data.pruneOneItemsCor();
			data.pruneOneItemsFromTrans();
		}
		
		Set<ArrayList<Integer>> keys = candidates.keySet();
		for(ArrayList<Integer> key:keys){
			if(data.pruningType==CorMiner.ADVANCED_PRUNING){
				boolean toPrune = false;
				for(int oneItem:key){
					if(!data.oneItemSets.containsKey(oneItem) || data.oneItemSets.get(oneItem).cor < data.corThr) {
						toPrune = true;
						break;
					}
				}
				if(toPrune) continue;
			}
			int lastElement = key.remove(key.size()-1);
			if(nextCands.containsKey(key)){
				HashMap<Integer,Object> vals = nextCands.get(key);
				vals.put(lastElement,null);
				nextCands.put(key, vals);
			} else {
				HashMap<Integer,Object> vals = new HashMap<Integer,Object>();
				vals.put(lastElement,null);
				nextCands.put(key, vals);
			}
		}
	}
	
	HashMap<ArrayList<Integer>,Info> generateNextCands(){
		HashMap<ArrayList<Integer>,Info> tempCands = new HashMap<ArrayList<Integer>,Info>();
		
		Set<ArrayList<Integer>> keys = nextCands.keySet();
		for(ArrayList<Integer> key:keys){
			HashMap<Integer,Object> vals = nextCands.get(key);
			Integer[] valsArr = vals.keySet().toArray(new Integer[0]);
			Arrays.sort(valsArr);
			int len = valsArr.length;
			for(int i=0;i<len;i++){
				for(int j=i+1;j<len;j++){
					ArrayList<Integer> prefix = new ArrayList<Integer>(key);
					prefix.add(valsArr[i]);
					prefix.add(valsArr[j]);
					tempCands.put(prefix, new Info());
				}
			}
		}
		
		return tempCands;
	}
	
}
