

import java.util.ArrayList;

public class CorMiner {
	
	public static int BASIC_PRUNING=0;		// Minsup pruning
	public static int SC_PRUNING=1;  		// BASIC + Stop Criteria (using correlation measure)
	public static int ADVANCED_PRUNING=2; 	// BASIC + SC + Pattern Generation pruning
	
	public static int GIVEN_CORTHR=0;
	public static int GIVEN_TOPKBASIC=1;
	public static int GIVEN_TOPKADVANCED=2;
	public String dictDelimiter=" ";
	int pruningType;
	
	public CorMiner(int pruningMethod){
		this.pruningType = pruningMethod;
	}
	
	// support counting of candidates generated by F_{k-1}*F_{k-1} method
	void run(String inputFile, double minsup, double corThr) throws Exception{
		Data data = new Data(pruningType);
long start=System.currentTimeMillis();
		data.readFile(inputFile); //Read input data and compute supports of 1-itemsets
long end=System.currentTimeMillis();
System.out.println((end-start)+" ms to read file.");
start=System.currentTimeMillis();
		data.setThresholds(minsup, corThr);
		System.out.println("Minsup: " + data.minsup);
		System.out.println("Cor Thr: " + corThr);
		data.compute1ItemSets();
//		System.out.println(data.oneItemSets.keySet());
		System.out.println(data.oneItemSets.size());
		
		int k=2;
//		ArrayList<Cell> table = new ArrayList<Cell>();
		Cell curCell = new Cell(data);
		curCell.candidates = data.generate2ItemSets();
		curCell.computeCandsFromCands(k);
		
		while(!curCell.candidates.isEmpty()){
			if(pruningType>=CorMiner.SC_PRUNING && curCell.allLessCorThr) break;
			
//			System.out.println(curCell.candidates.keySet());
			for(ArrayList<Integer> key:curCell.candidates.keySet()){
				Info in = curCell.candidates.get(key);
				System.out.println(key+": ("+in.sup+","+in.cor+")");
			}
			System.out.println(curCell.candidates.size());
			curCell.prepareNextCands();
			Cell cell = new Cell(data);
			cell.candidates = curCell.generateNextCands();
			k++;
			cell.computeCandsFromCands(k);
			curCell = cell;
		}
		System.out.println("Number of solutions: "+data.sols.size());
System.out.println((System.currentTimeMillis()-start)+" ms to finish Sang computation.");
		data.printSol();
	}

	// support counting with generating patterns in each transaction
	void run1(String inputFile, double minsup, double corThr) throws Exception{
		Data data = new Data(pruningType);
long start=System.currentTimeMillis();
		data.readFile(inputFile); //Read input data and compute supports of 1-itemsets
long end=System.currentTimeMillis();
System.out.println((end-start)+" ms to read file.");
start=System.currentTimeMillis();
		data.setThresholds(minsup, corThr);
		System.out.println("Minsup: " + data.minsup);
		System.out.println("Cor Thr: " + corThr);
		data.compute1ItemSets();
//		System.out.println(data.oneItemSets.keySet());
		System.out.println(data.oneItemSets.size());
		
		int k=2;
//		ArrayList<Cell> table = new ArrayList<Cell>();
		Cell curCell = new Cell(data);
		curCell.candidates = data.generate2ItemSets();
		curCell.computeCandsFromTransNew(k);
		
		while(!curCell.candidates.isEmpty()){
			if(pruningType>=CorMiner.SC_PRUNING && curCell.allLessCorThr) break;
			 
//			System.out.println(curCell.candidates.keySet());
			System.out.println(curCell.candidates.size());
			curCell.prepareNextCands();
			Cell cell = new Cell(data);
			cell.candidates = curCell.generateNextCands();
			k++;
			cell.computeCandsFromTransNew(k);
			curCell = cell;
		}
		System.out.println("Number of solutions: "+data.sols.size());
System.out.println((System.currentTimeMillis()-start)+" ms to finish Marina computation.");
		data.printSol();
	}
	
	// support counting in mixed way
	void run2(String inputFile, double minsup, double corThr,String dictFileName) throws Exception{
		Data data = new Data(pruningType);
long start=System.currentTimeMillis();
		data.readFile(inputFile); //Read input data and compute supports of 1-itemsets
long end=System.currentTimeMillis();
System.out.println((end-start)+" ms to read file.");
start=System.currentTimeMillis();
		data.setThresholds(minsup, corThr);
		System.out.println("Minsup: " + data.minsup);
		System.out.println("Cor Thr: " + corThr);
		data.compute1ItemSets();
//		System.out.println(data.oneItemSets.keySet());
		System.out.println(data.oneItemSets.size());
		
		int k=2;
//		ArrayList<Cell> table = new ArrayList<Cell>();
		Cell curCell = new Cell(data);
		curCell.candidates = data.generate2ItemSets();
		//System.out.print("<"+k+"-itemsets generated from Trans> ");
		//curCell.computeCandsFromTrans(k);
		if(comb(data.aveTranLength,k)>curCell.candidates.size()){
			System.out.print("<"+k+"-itemsets matched from Cands> ");
			curCell.computeCandsFromCands(k);
		} else {
			System.out.print("<"+k+"-itemsets matched from Trans> ");
			curCell.computeCandsFromTransNew(k);
		}
		
		while(!curCell.candidates.isEmpty()){
			if(pruningType>=CorMiner.SC_PRUNING && curCell.allLessCorThr){
				System.out.println("Stopped due to allLessCorThr");
				break;
			}
			 
//			System.out.println(curCell.candidates.keySet());
			System.out.println(curCell.candidates.size());
			curCell.prepareNextCands();
			Cell cell = new Cell(data);
			cell.candidates = curCell.generateNextCands();
			k++;
			if(comb(data.aveTranLength,k)>cell.candidates.size()){
				System.out.print("<"+k+"-itemsets matched from Cands> ");
				cell.computeCandsFromCands(k);
			} else {
				System.out.print("<"+k+"-itemsets matched from Trans> ");
				cell.computeCandsFromTransNew(k);
			}
			curCell = cell;
		}
		System.out.println();
		System.out.println("Number of solutions: "+data.sols.size());
System.out.println((System.currentTimeMillis()-start)+" ms to finish Mixed computation.");
		data.printSol(dictFileName,dictDelimiter);
	}
	
	int comb(int total, int k){
		int val = 1;
		int until = total-k+1;
		for(int i=total;i>=until;i--){
			val*=i;
		}
		for(int i=k;i>0;i--){
			val/=i;
		}
		return val;
	}
	
	// mining top-k correlation patterns
	void runTopKAdvanced(String inputFile, double minsup, int topK, String dictFileName) throws Exception{
		Data data = new Data(pruningType);
long start=System.currentTimeMillis();
		data.readFile(inputFile); //Read input data and compute supports of 1-itemsets
long end=System.currentTimeMillis();
System.out.println((end-start)+" ms to read file.");
start=System.currentTimeMillis();
		data.setThresholds(minsup, topK);
		System.out.println("Total transaction: " + (minsup*data.minsup));
		System.out.println("Minsup abs: " + data.minsup+ " relative: "+minsup);
		System.out.println("Top-K: " + topK);
		data.compute1ItemSets();
//		System.out.println(data.oneItemSets.keySet());
		System.out.println(data.oneItemSets.size());
		
		int k=2;
//		ArrayList<Cell> table = new ArrayList<Cell>();
		Cell curCell = new Cell(data);
		curCell.candidates = data.generate2ItemSets();
		//System.out.print("<"+k+"-itemsets matched from Trans> ");
		//curCell.computeCandsTopKFromTrans(k);
		if(comb(data.aveTranLength,k)>curCell.candidates.size()){
			System.out.print("<"+k+"-itemsets matched from Cands> ");
			curCell.computeCandsTopKFromCands(k);
		} else {
			System.out.print("<"+k+"-itemsets matched from Trans> ");
			curCell.computeCandsTopKFromTransNew(k);
		}
		
		while(!curCell.candidates.isEmpty()){
			if(pruningType>=CorMiner.SC_PRUNING && curCell.allLessCorThr) {
				System.out.println("Stopped due to allLessCorThr");
				break;
			}
			 
//			System.out.println(curCell.candidates.keySet());
			System.out.println(curCell.candidates.size());
			curCell.prepareNextCands();
			Cell cell = new Cell(data);
			cell.candidates = curCell.generateNextCands();
			k++;
			if(comb(data.aveTranLength,k)>cell.candidates.size()){
				System.out.print("<"+k+"-itemsets matched from Cands> ");
				cell.computeCandsTopKFromCands(k);
			} else {
				System.out.print("<"+k+"-itemsets matched from Trans> ");
				cell.computeCandsTopKFromTransNew(k);
			}
			curCell = cell;
		}
		System.out.println();
		System.out.println("Number of solutions: "+data.sols.size());
		System.out.println("correlation threshold: "+data.corThr);
System.out.println((System.currentTimeMillis()-start)+" ms to finish top-k computation.");
		data.printSol(dictFileName,dictDelimiter);
//		int count=0;
//		for(double key:data.topkCors.keySet()){
//			count += data.topkCors.get(key);
//		}
//		System.out.println("Num of topkCors: "+count);
//		count=0;
//		for(double key:data.topkInvs.keySet()){
//			count += data.topkInvs.get(key).size();
//		}
//		System.out.println("Num of topkInvs: "+count);
	}
	
	// mining top-k correlation patterns in an iterative way
	void runTopKBasic(String inputFile, double minsup, int topK,String dictFileName) throws Exception{
long start=System.currentTimeMillis();

start=System.currentTimeMillis();
		double corThr = 1;
		System.out.println("Top-K: " + topK);
		while(!run2ForTopk(inputFile,minsup,corThr,topK,dictFileName)){
			corThr /= 2;
		}
System.out.println((System.currentTimeMillis()-start)+" ms to finish top-k basic computation.");
	}
	
	// support counting in mixed way
	boolean run2ForTopk(String inputFile, double minsup, double corThr, int topK,String dictFileName) throws Exception{
		Data data = new Data(pruningType);
long start=System.currentTimeMillis();
		data.readFile(inputFile); //Read input data and compute supports of 1-itemsets
long end=System.currentTimeMillis();
System.out.println((end-start)+" ms to read file.");
start=System.currentTimeMillis();
		data.setThresholds(minsup, corThr);
		System.out.println("Minsup: " + data.minsup);
		System.out.println("Cor Thr: " + corThr);
		data.compute1ItemSets();
//		System.out.println(data.oneItemSets.keySet());
		System.out.println(data.oneItemSets.size());
		
		int k=2;
//		ArrayList<Cell> table = new ArrayList<Cell>();
		Cell curCell = new Cell(data);
		curCell.candidates = data.generate2ItemSets();
		//System.out.print("<"+k+"-itemsets matched from Trans> ");
		//curCell.computeCandsFromTrans(k);
		if(comb(data.aveTranLength,k)>curCell.candidates.size()){
			System.out.print("<"+k+"-itemsets matched from Cands> ");
			curCell.computeCandsFromCands(k);
		} else {
			System.out.print("<"+k+"-itemsets matched from Trans> ");
			curCell.computeCandsFromTransNew(k);
		}
		
		while(!curCell.candidates.isEmpty()){
			if(pruningType>=CorMiner.SC_PRUNING && curCell.allLessCorThr){
				System.out.println("Stopped due to allLessCorThr");
				break;
			}
			 
//			System.out.println(curCell.candidates.keySet());
			System.out.println(curCell.candidates.size());
			curCell.prepareNextCands();
			Cell cell = new Cell(data);
			cell.candidates = curCell.generateNextCands();
			k++;
			if(comb(data.aveTranLength,k)>cell.candidates.size()){
				System.out.print("<"+k+"-itemsets matched from Cands> ");
				cell.computeCandsFromCands(k);
			} else {
				System.out.print("<"+k+"-itemsets matched from Trans> ");
				cell.computeCandsFromTransNew(k);
			}
			curCell = cell;
		}
		int nSol = data.sols.size();
		System.out.println();
		System.out.println("Number of solutions: "+nSol);
System.out.println((System.currentTimeMillis()-start)+" ms to finish Mixed computation.");
	data.printSol(dictFileName,dictDelimiter);
		return (topK<=nSol);
	}
	
	
	public static void main(String[] args) throws Exception{
		
		String dictDelim=" ";
		String inputFile="NSFawards/wordtransactions.txt";
		//String inputFile="data/test/test2.txt";
		String dictionaryFile="NSFawards/words.txt";
		double minsup=0.01;
		double corThr=0.3;
		int topK=10;
		int pruningMethod = CorMiner.ADVANCED_PRUNING;
		//int pruningMethod = CorMiner.SC_PRUNING;
		//int pruningMethod = CorMiner.BASIC_PRUNING;
		
		//int miningType =CorMiner.GIVEN_TOPKBASIC;
		//int miningType = CorMiner.GIVEN_TOPKADVANCED;
		int miningType = CorMiner.GIVEN_CORTHR;
/*		
		if(args.length<5){
			System.err.println("Insufficient number of arguments");
			System.err.println("java CorMiner [Mining Type(0-2)] [minsup] [Cor Threshold/Top-k] [Pruning Method(0-2)] [Input File]");
			System.exit(1);
		}
		
		miningType = Integer.parseInt(args[0]);
		minsup = Double.parseDouble(args[1]);
		pruningMethod = Integer.parseInt(args[3]);
		inputFile = args[4];
*/
		CorMiner cm = new CorMiner(pruningMethod);
		cm.dictDelimiter=dictDelim;
		if(miningType==CorMiner.GIVEN_CORTHR){
			cm.run2(inputFile,minsup,corThr,dictionaryFile);
		} else if(miningType==CorMiner.GIVEN_TOPKBASIC){
			cm.runTopKBasic(inputFile,minsup,topK,dictionaryFile);
		} else if(miningType==CorMiner.GIVEN_TOPKADVANCED){
			cm.runTopKAdvanced(inputFile,minsup,topK,dictionaryFile);
		}
/*	
		CorMiner cm = new CorMiner(pruningMethod);
//		cm.run(inputFile,minsup,corThr);
//		System.out.println();
//		cm.run1(inputFile,minsup,corThr);
//		System.out.println();
//		cm.run2(inputFile,minsup,corThr);
//		System.out.println();
		cm.runTopKAdvanced(inputFile,minsup,topK);
//		cm.runTopKBasic(inputFile,minsup,topK);
*/

	}

}
