
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeMap;

public class Data {

	ArrayList<HashMap<Integer,Object>> trans;
	int size;
	int aveTranLength;
	int minsup;
	double corThr;
	int topK;
	int pruningType;
	HashMap<Integer,Info> oneItemSets;
	HashMap<ArrayList<Integer>,Info> sols;
	TreeMap<Double,Integer> topkCors;	// top-k correlation values
	HashMap<Double,HashMap<ArrayList<Integer>,Object>> topkInvs; // inverse mapping from top-k correlation values to their corresponding patterns
	
	
	public Data(int pruningType){
		trans = new ArrayList<HashMap<Integer,Object>>();
		oneItemSets = new HashMap<Integer,Info>();
		sols = new HashMap<ArrayList<Integer>,Info>();
		this.pruningType = pruningType;
		aveTranLength = 0;
		topkCors = new TreeMap<Double,Integer>();
		topkInvs = new HashMap<Double,HashMap<ArrayList<Integer>,Object>>();
	}
	
	void readFile(String inputFile) throws Exception{
		BufferedReader rdr = new BufferedReader(new InputStreamReader(new FileInputStream(inputFile), "UTF-8"));
		String line;
		while((line=rdr.readLine())!=null){
//			System.out.println(line);
			String[] items = line.split("\\s+");
			aveTranLength += items.length;
			HashMap<Integer,Object> tran = new HashMap<Integer,Object>();
			try{
			for(String item:items){
				int itemN = Integer.parseInt(item);
				tran.put(itemN,null);
				if(oneItemSets.containsKey(itemN)){
					Info info = oneItemSets.get(itemN);
					info.sup++;
					oneItemSets.put(itemN, info);
				} else {
					Info info = new Info();
					info.sup++;
					oneItemSets.put(itemN, info);
				}
			}
			}catch(Exception e){
				System.err.println("line: "+line);
				throw e;
			}
			trans.add(tran);
		}
		rdr.close();
		size = trans.size();
		aveTranLength = aveTranLength/size;
		System.out.println("Average transaction width: "+aveTranLength);
	}
	
	void setThresholds(double minsup, double corThr){
		this.minsup = (int)(minsup*size);
		this.corThr = corThr;
	}
	
	void setThresholds(double minsup, int topK){
		this.minsup = (int)(minsup*size);
		this.corThr = 0;
		this.topK = topK;
	}
	
	void compute1ItemSets(){
		//Apply minsup
		Set<Integer> keys = oneItemSets.keySet();
		HashMap<Integer,Object> toRemove = new HashMap<Integer,Object>();
		System.out.println("Different items: "+oneItemSets.size());
		for(int item:keys){
			Info info = oneItemSets.get(item);
			if(info.sup < minsup) toRemove.put(item,null);
		}
		Set<Integer> items = toRemove.keySet();
		for(int item:items) oneItemSets.remove(item);
	}
	
	HashMap<ArrayList<Integer>,Info> generate2ItemSets(){
		HashMap<ArrayList<Integer>,Info> tempCands = new HashMap<ArrayList<Integer>,Info>();
		
		//Set<Integer> keys = oneItemSets.keySet();
		Integer[] keysArr = oneItemSets.keySet().toArray(new Integer[0]);
		Arrays.sort(keysArr);
		int len = keysArr.length;
		for(int i=0;i<len;i++){
			for(int j=i+1;j<len;j++){
				ArrayList<Integer> twoItems = new ArrayList<Integer>();
				twoItems.add(keysArr[i]);
				twoItems.add(keysArr[j]);
				tempCands.put(twoItems, new Info());
			}
		}
		
		return tempCands;
	}

//	void initOneItemsCor(){
//		Set<Integer> keys = oneItemSets.keySet();
//		
//		for(int item:keys){
//			Info info = oneItemSets.get(item);
//			info.cor = 0;
//			oneItemSets.put(item, info);
//		}
//	}

	void initOneItemsCor(){
		Set<Integer> keys = oneItemSets.keySet();
		
		for(int item:keys){
			Info info = oneItemSets.get(item);
			//System.out.println(item+":"+info.sup+","+info.cor);
			info.cor = 0;
			oneItemSets.put(item, info);
		}
	}

	void pruneOneItemsCor(){
		Set<Integer> keys = oneItemSets.keySet();
		HashMap<Integer,Object> toRemove = new HashMap<Integer,Object>();
		
		for(int item:keys){
			Info info = oneItemSets.get(item);
			if(info.cor < corThr) toRemove.put(item,null);
		}
		Set<Integer> items = toRemove.keySet();
		for(int item:items) oneItemSets.remove(item);
	}
	
	void pruneOneItemsCorSorted(){
		Set<Integer> keys = oneItemSets.keySet();
		HashMap<Integer,Object> toRemove = new HashMap<Integer,Object>();
		
		TreeMap<Double,HashSet<Integer>> sortedItems = new TreeMap<Double,HashSet<Integer>>();
		for(int item:keys){
			Info info = oneItemSets.get(item);
			if(sortedItems.containsKey(info.cor)){
				HashSet<Integer> items = sortedItems.get(info.cor);
				items.add(item);
				sortedItems.put(info.cor, items);
			} else {
				HashSet<Integer> items = new HashSet<Integer>();
				items.add(item);
				sortedItems.put(info.cor, items);
			}
		}
		
		for(double cor:sortedItems.keySet()){
//			boolean toContinue = true;
			HashSet<Integer> items = sortedItems.get(cor);
			
//			for(int item:items){
//				Info info = oneItemSets.get(item);
//				if(info.cor < corThr) toRemove.put(item,null);
//				else {
//					toContinue = false;
//					break;
//				}
//				if(!toContinue) break;
//			}
			if(cor < corThr){
				for(int item:items){
					toRemove.put(item,null);
				}
			} else break;
		}
		
		Set<Integer> items = toRemove.keySet();
		for(int item:items) oneItemSets.remove(item);
	}
	
	void pruneOneItemsFromTrans(){
//		aveTranLength = 0;
		for(HashMap<Integer,Object> tran:trans){
			HashMap<Integer,Object> toRemove = new HashMap<Integer,Object>();
			Set<Integer> row = tran.keySet();
			for(int item:row){
				if(!oneItemSets.containsKey(item)) toRemove.put(item, null);
			}
			Set<Integer> removeItems = toRemove.keySet();
			for(int removeItem:removeItems) tran.remove(removeItem);
//			int tSize = tran.size();
//			if(tSize>0)	aveTranLength += tran.size();
//			else trans.remove(tran);
		}
//		size = trans.size();
//		aveTranLength = aveTranLength/size;
	}
	
	void printSol(){
//		Set<ArrayList<Integer>> keys = sols.keySet();
//		for(ArrayList<Integer> key:keys){
		HashSet<Double> cors = new HashSet<Double>();
		HashMap<Double,HashSet<ArrayList<Integer>>> invCors = new HashMap<Double,HashSet<ArrayList<Integer>>>();
		Set<ArrayList<Integer>> keys = sols.keySet();
		for(ArrayList<Integer> key:keys){
			Info info = sols.get(key);
			cors.add(info.cor);
			if(invCors.containsKey(info.cor)){
				HashSet<ArrayList<Integer>> invKeys = invCors.get(info.cor);
				invKeys.add(key);
				invCors.put(info.cor, invKeys);
			} else {
				HashSet<ArrayList<Integer>> invKeys = new HashSet<ArrayList<Integer>>();
				invKeys.add(key);
				invCors.put(info.cor, invKeys);
			}
		}
		
		Double[] sortedCors = cors.toArray(new Double[0]);
		Arrays.sort(sortedCors, Collections.reverseOrder());
		ArrayList<ArrayList<Integer>> sortedKeys = new ArrayList<ArrayList<Integer>>();
		for(double cor:sortedCors){
			HashSet<ArrayList<Integer>> invCor = invCors.get(cor);
			for(ArrayList<Integer> key:invCor){
				sortedKeys.add(key);
			}
		}
		
//		Iterator<ArrayList<Integer>> ite = sols.keySet().iterator();
//		while(ite.hasNext()){
//			ArrayList<Integer> key = ite.next();
		for(ArrayList<Integer> key:sortedKeys){
			if(sols.get(key)==null) System.out.println(key+" null info");
			else {
				//System.out.print(key+" sup:"+sols.get(key).sup+" cor:"+sols.get(key).cor);
				System.out.printf("%s sup:%d cor:%.3f",key.toString(),sols.get(key).sup,sols.get(key).cor);
				
				for(int i=0;i<key.size();i++){
					int item = key.get(i);
					System.out.printf(" %d:%d", item, this.oneItemSets.get(item).sup);
				}
				System.out.println();
			}
		}
		System.out.println();
//		Collection<Info> infos = sols.values();
//		for(Info info:infos){
//			System.out.println(info.cor);
//		}
	}
	
	void printSol(String dictionaryName, String delim)
	{
		Map <Integer,String> dictionary=new HashMap <Integer,String>();
		try
		{
			FileInputStream fstream = new FileInputStream(dictionaryName);
		    DataInputStream dis = new DataInputStream(fstream);
		    BufferedReader in= new BufferedReader(new InputStreamReader(dis));
		    String line;
		    while((line=in.readLine())!=null)
		    {
		    	StringTokenizer st=new StringTokenizer(line,delim);	
		    	Integer key=Integer.parseInt(st.nextToken());
		    	String val=st.nextToken();
		    	dictionary.put(key, val);
		    }
		    in.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
			return;
		}
		HashSet<Double> cors = new HashSet<Double>();
		HashMap<Double,HashSet<ArrayList<Integer>>> invCors = new HashMap<Double,HashSet<ArrayList<Integer>>>();
		Set<ArrayList<Integer>> keys = sols.keySet();
		for(ArrayList<Integer> key:keys){
			Info info = sols.get(key);
			cors.add(info.cor);
			if(invCors.containsKey(info.cor)){
				HashSet<ArrayList<Integer>> invKeys = invCors.get(info.cor);
				invKeys.add(key);
				invCors.put(info.cor, invKeys);
			} else {
				HashSet<ArrayList<Integer>> invKeys = new HashSet<ArrayList<Integer>>();
				invKeys.add(key);
				invCors.put(info.cor, invKeys);
			}
		}
		
		Double[] sortedCors = cors.toArray(new Double[0]);
		Arrays.sort(sortedCors, Collections.reverseOrder());
		ArrayList<ArrayList<Integer>> sortedKeys = new ArrayList<ArrayList<Integer>>();
		for(double cor:sortedCors){
			HashSet<ArrayList<Integer>> invCor = invCors.get(cor);
			for(ArrayList<Integer> key:invCor){
				sortedKeys.add(key);
			}
		}
		
		for(ArrayList<Integer> key:sortedKeys){

			if(sols.get(key)==null) System.out.println(key+" null info");
			else 
			{
				//if(key.size()>2)
				//{
				ArrayList <String> captionsList=convertToCaptions(dictionary,key);
				//System.out.print(captionsList+" sup:"+sols.get(key).sup+" cor:"+sols.get(key).cor);
				System.out.printf("%s sup:%d cor:%.3f",captionsList.toString(),sols.get(key).sup,sols.get(key).cor);
				for(int i=0;i<key.size();i++){
					int item = key.get(i);
					if(oneItemSets.get(item)!= null)
					{
						System.out.printf(" sup(%s):%d", dictionary.get(item), oneItemSets.get(item).sup);
					}
				}
				System.out.println();
				//}
			}
		}
		System.out.println();
	}
	
	private static ArrayList <String> convertToCaptions(Map <Integer,String> dictionary,ArrayList<Integer> itemset )
	{
		ArrayList <String> ret=new ArrayList <String>();
		for(int i=0;i<itemset.size();i++)
			ret.add(dictionary.get(itemset.get(i)));
		
		return ret;
	}

}
